/**
 * Copyright 2009-15-13 Simon Andrews
 *
 *    This file is part of SeqMonk.
 *
 *    SeqMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SeqMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with SeqMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.babraham.SeqMonk.Displays.DistanceValuePlot;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import uk.ac.babraham.SeqMonk.SeqMonkException;
import uk.ac.babraham.SeqMonk.DataTypes.DataStore;
import uk.ac.babraham.SeqMonk.DataTypes.Genome.Chromosome;
import uk.ac.babraham.SeqMonk.DataTypes.Probes.Probe;
import uk.ac.babraham.SeqMonk.DataTypes.Probes.ProbeList;
import uk.ac.babraham.SeqMonk.Dialogs.CrashReporter;
import uk.ac.babraham.SeqMonk.Preferences.ColourScheme;
import uk.ac.babraham.SeqMonk.Utilities.AxisScale;

/**
 * The DistanceValue Panel works somewhat like a histogram, except that
 * it displays average quantitation values for probes which fall a 
 * certain distance from a target.
 */
public class DistanceValuePanel extends JPanel implements Runnable, ChangeListener {

	/** The data. */
	private DataStore data;
	private ProbeList allProbes;
	private ProbeList targetProbes;
	private int maxDistance;
	
	private DistanceValue [] distanceValues;
	
	/** The category slider. */
	private JSlider categorySlider;
	
	/** The main histogram panel. */
	private MainHistogramPanel mainHistogramPanel;
	
	/** The status panel. */
	private StatusPanel statusPanel;
	
	/** The Constant df. */
	private static final DecimalFormat df = new DecimalFormat("#.##");
	
	/** The calculating categories. */
	private boolean calculatingCategories = false;
	
	/** The stop calculating. */
	private boolean stopCalculating = false;
	
	/** The min data value. */
	private double minDataValue;
	
	/** The max data value. */
	private double maxDataValue;
	
	/** The current category count. */
	private int currentCategoryCount = 0;
	
	/** The currently displayed min value after zooming */
	private double currentMinValue;
	
	/** The currently displayed max value after zooming */
	private double currentMaxValue;
	
	/* These values are used when dragging a selection to zoom */
	private int selectionStart;
	private int selectionEnd;
	private boolean isSelecting = false;
	
	
	
	/**
	 * Instantiates a new histogram panel.
	 * 
	 * @param data the data
	 */
	public DistanceValuePanel (DataStore data, ProbeList allProbes, ProbeList targetProbes, int maxDistance, boolean negativeDistances) {
		
		this.data = data;
		this.maxDistance = maxDistance;
		this.allProbes = allProbes;
		this.targetProbes = targetProbes;
		
		// We need to find the min and max values for the
		// data.  We can do this here since it won't change
	
		if (negativeDistances) {
			minDataValue = 0-maxDistance;
		}
		else {
			minDataValue = 0;
		}
		maxDataValue = maxDistance;
		
		try {
			distanceValues = calculateDistanceValues();
		} 
		catch (SeqMonkException e1) {
			new CrashReporter(e1);
		}
		
		// We initially default to showing the whole dataset
		currentMinValue = minDataValue;
		currentMaxValue = maxDataValue;
		
		setLayout(new BorderLayout());
				
		JPanel topPanel = new JPanel();
		topPanel.setLayout(new BorderLayout());
		categorySlider = new JSlider(5,200,25);
		categorySlider.setMajorTickSpacing(15);
		categorySlider.setPaintTicks(true);
		categorySlider.setPaintLabels(true);
		categorySlider.addChangeListener(this);
		topPanel.add(categorySlider,BorderLayout.CENTER);
		topPanel.add(new JLabel("Divisions"),BorderLayout.WEST);
				
		add(topPanel,BorderLayout.NORTH);
		
		
		mainHistogramPanel = new MainHistogramPanel();
		add(mainHistogramPanel,BorderLayout.CENTER);
		
		statusPanel = new StatusPanel();
		add(statusPanel,BorderLayout.SOUTH);
		
		calcuateCategories(categorySlider.getValue());
		
	}
	
	/**
	 * Main histogram panel.
	 * 
	 * @return the j panel
	 */
	public JPanel mainHistogramPanel () {
		return mainHistogramPanel;
	}
	
	/* (non-Javadoc)
	 * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
	 */
	public void stateChanged(ChangeEvent ce) {
		// When this fires it means that the categorySlider
		// has moved so we need to recalculate the graph.
				
		calcuateCategories(categorySlider.getValue());
	}
	
	/**
	 * Calcuate categories.
	 * 
	 * @param categoryCount the category count
	 */
	private void calcuateCategories (int categoryCount) {
		
		// If we're already calculating then stop and do it
		// again with the new value
				
		if (calculatingCategories) {
//			System.out.println("Waiting for previous calculation to finish");
			stopCalculating = true;
			while (calculatingCategories) {
//				System.out.println("Still waiting");
				try {
					Thread.sleep(50);
				} 
				catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}

		currentCategoryCount = categoryCount;
		
		Thread t = new Thread(this);
		t.start();
		
	}
	
	private DistanceValue [] calculateDistanceValues () throws SeqMonkException {
		Vector<DistanceValue> values = new Vector<DistanceValue>();
		
		Probe [] targetProbes = this.targetProbes.getAllProbes();
		
		Probe [] allChrProbes = new Probe[0];
		
		Chromosome lastChr = null;
		int lastStart = 0;
		
		for (int p=0;p<targetProbes.length;p++) {
			
			if (targetProbes[p].chromosome() != lastChr) {
				allChrProbes = allProbes.getProbesForChromosome(targetProbes[p].chromosome());
				lastChr = targetProbes[p].chromosome();
				lastStart = 0;
			}
			
			for (int i=lastStart;i<allChrProbes.length;i++) {
				if (allChrProbes[i].middle() < targetProbes[p].middle()-maxDistance) {
					lastStart = i;
					continue;
				}
				else if (allChrProbes[i].middle() > targetProbes[p].middle()+maxDistance) {
					break;
				}
				values.add(new DistanceValue(allChrProbes[i].middle()-targetProbes[p].middle(), data.getValueForProbe(allChrProbes[i])));
			}
			
		}
		
		
		DistanceValue [] returnArray = values.toArray(new DistanceValue[0]);
		Arrays.sort(returnArray);
		return returnArray;
	}
	
	
	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

//		System.out.println("Calculating "+currentCategoryCount+" categories");
		
		calculatingCategories = true;
		
		// In this thread we recalculate the categories to display
		
		HistogramCategory [] categories = new HistogramCategory[currentCategoryCount];
		
		double interval = (currentMaxValue-currentMinValue)/categories.length;
		
		for (int c=0;c<categories.length;c++) {
			categories[c] = new HistogramCategory(currentMinValue+(interval*c),currentMinValue+(interval*(c+1)));
//			System.out.println("Made category from "+categories[c].minValue+"-"+categories[c].maxValue);
		}
		
		// Go through both the categories and the distance values to find the average
		// value for each category.
		
		int lastDistanceValueIndex = 0;
		
		for (int c=0;c<categories.length;c++) {

			for (int v=lastDistanceValueIndex;v<distanceValues.length;v++) {
				
				if (stopCalculating) {
					stopCalculating = false;
					calculatingCategories = false;
					return;
				}
				
				if (distanceValues[v].distance < categories[c].minValue) {
					lastDistanceValueIndex = v;
					continue;
				}
				else if (distanceValues[v].distance > categories[c].maxValue) {
					break;
				}
				else {
					categories[c].addValue(distanceValues[v].value);
				}
			}
			
		}
		
		
		mainHistogramPanel.setData(categories);
		
		calculatingCategories = false;
//		System.out.println("Finished counting");
		
	}
	
	/**
	 * The Class MainHistogramPanel.
	 */
	private class MainHistogramPanel extends JPanel implements MouseListener,MouseMotionListener {
		
		private static final int X_AXIS_SPACE = 50;
		private static final int Y_AXIS_SPACE = 30;
		
		
		/** The categories. */
		private HistogramCategory [] categories = new HistogramCategory[0];
		
		/** The selected category. */
		private HistogramCategory selectedCategory = null;
		
		/** The max count. */
		private float maxCount;
		
		/**
		 * Instantiates a new main histogram panel.
		 */
		public MainHistogramPanel () {
			addMouseListener(this);
			addMouseMotionListener(this);
		}
		
		/**
		 * Sets the data.
		 * 
		 * @param categories the new data
		 */
		public void setData (HistogramCategory [] categories) {
			this.categories = categories;
			maxCount = 0;
			for (int c=0;c<categories.length;c++) {
				if (categories[c].value() > maxCount) maxCount = categories[c].value();
			}
			
			repaint();
		}
		
		/* (non-Javadoc)
		 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
		 */
		public void paintComponent (Graphics g) {
			super.paintComponent(g);
			
			// We want a white background
			g.setColor(Color.WHITE);
			g.fillRect(0, 0, getWidth(), getHeight());
			
			// Draw the graph axes first.  We leave a border on all sides
			g.setColor(Color.BLACK);
			
			g.drawLine(X_AXIS_SPACE, 5, X_AXIS_SPACE, getHeight()-Y_AXIS_SPACE);
			g.drawLine(X_AXIS_SPACE, getHeight()-Y_AXIS_SPACE, getWidth()-5, getHeight()-Y_AXIS_SPACE);
			
			// If we don't have any data we can stop here
			if (categories == null) return;
			
			// If we're making a selection then draw the selected region
			if (isSelecting) {
				g.setColor(ColourScheme.DRAGGED_SELECTION);
				g.fillRect(Math.min(selectionStart,selectionEnd), 5, Math.abs(selectionStart-selectionEnd), getHeight()-(Y_AXIS_SPACE+5));
				g.setColor(Color.BLACK);
			}
			

			// We need the scaling factor for the y-axis
			double yScale = 0;
			yScale = (double)(getHeight()-(5+Y_AXIS_SPACE))/maxCount;
			
			
			// Now draw the scale on the y axis
			AxisScale yAxisScale = new AxisScale(0, maxCount);
			
			double currentYValue = yAxisScale.getStartingValue();
			
			while (currentYValue < maxCount) {
				
				double yHeight = 0;
				yHeight = currentYValue*yScale;
				
				g.drawString(yAxisScale.format(currentYValue), 2, (int)((getHeight()-Y_AXIS_SPACE)-yHeight)+(g.getFontMetrics().getAscent()/2));

				// Put a line across the plot
				if (currentYValue != 0) {
					g.setColor(Color.LIGHT_GRAY);
					g.drawLine(X_AXIS_SPACE, (int)((getHeight()-Y_AXIS_SPACE)-yHeight), getWidth()-5, (int)((getHeight()-Y_AXIS_SPACE)-yHeight));
					g.setColor(Color.BLACK);
				}
				
				currentYValue += yAxisScale.getInterval();
			}
			
			// Now draw the scale on the x axis
			if (categories.length>0) {
				AxisScale xAxisScale = new AxisScale(currentMinValue, currentMaxValue);
								
				double currentXValue = xAxisScale.getStartingValue();
				while (currentXValue < currentMaxValue) {
					g.drawString(xAxisScale.format(currentXValue), getX(currentXValue), (int)((getHeight()-Y_AXIS_SPACE)+15));
					
					currentXValue += xAxisScale.getInterval();
					
				}
			}
			
			
			
			// Now we can draw the different categories
			for (int c=0;c<categories.length;c++) {
				categories[c].xStart = getX(categories[c].minValue);;
				categories[c].xEnd = getX(categories[c].maxValue);
				
				if (categories[c] == selectedCategory) {
					g.setColor(ColourScheme.HIGHLIGHTED_HISTOGRAM_BAR);
				}
				else {
					g.setColor(ColourScheme.HISTOGRAM_BAR);
				}
				
				double ySize = 0;
				ySize = categories[c].value()*yScale;
			
				g.fillRect(categories[c].xStart, (int)((getHeight()-Y_AXIS_SPACE)-ySize), categories[c].xEnd-categories[c].xStart, (int)(ySize));
				
				
				// Draw a box around it
				g.setColor(Color.BLACK);
				g.drawRect(categories[c].xStart, (int)((getHeight()-Y_AXIS_SPACE)-ySize), categories[c].xEnd-categories[c].xStart, (int)(ySize));
				
			}
			
		}
		
		public int getX (double xValue) {
			return X_AXIS_SPACE + (int)(((getWidth()-(X_AXIS_SPACE+5))/(currentMaxValue-currentMinValue))*(xValue-currentMinValue));
		}
		
		public double getXValue (int xPosition) {			
			return ((((double)(xPosition-X_AXIS_SPACE))/(getWidth()-(X_AXIS_SPACE+5)))*(currentMaxValue-currentMinValue))+currentMinValue;			
		}
		
		
		/* (non-Javadoc)
		 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
		 */
		public void mouseDragged(MouseEvent me) {
			if (isSelecting) {
				int x=me.getX();
				if (x<X_AXIS_SPACE) x = X_AXIS_SPACE;
				if (x>getWidth()-5) x = getWidth()-5;
				selectionEnd = x;
				repaint();
			}
		}

		/* (non-Javadoc)
		 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
		 */
		public void mouseMoved(MouseEvent me) {

			// If we're outside the main plot area we don't need to worry about it
			if (me.getX() < 5 || me.getX() > getWidth()-5 || me.getY()<5 || me.getY() > getHeight()-5) {
				if (selectedCategory != null) {
					selectedCategory = null;
					statusPanel.setSelectedCategory(null);
					repaint();
				}
				
				return;
			}
			
			// Check to see if we're in one of the categories
			for (int c=0;c<categories.length;c++) {
				if (me.getX() >= categories[c].xStart && me.getX()<=categories[c].xEnd) {
					// We don't worry about being inside the bar on the y-axis
					
					if (categories[c] != selectedCategory) {
						selectedCategory = categories[c];
						statusPanel.setSelectedCategory(selectedCategory);
						repaint();
					}
					
					return;
				}
			}
			
			
		}
		
		/* (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent arg0) {}

		/* (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent arg0) {}

		/* (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent arg0) {
			selectedCategory = null;
			statusPanel.setSelectedCategory(null);
			repaint();
		}

		/* (non-Javadoc)
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent me) {
			
			// Don't do anything if they pressed the right mouse button
			if ((me.getModifiers() & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK) {
				return;
			}

			// If they're inside the plot area then start a selection
			if (getXValue(me.getX())>=currentMinValue && getXValue(me.getX())<= currentMaxValue) {
				selectionStart = me.getX();
				isSelecting = true;
			}
			
			
		}

		/* (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent me) {
			
			// Zoom out if they pressed the right mouse button
			if ((me.getModifiers() & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK) {
				double interval = currentMaxValue-currentMinValue;
				double midPoint = currentMinValue+(interval/2);
				
				currentMinValue = Math.max(midPoint-interval, minDataValue);
				currentMaxValue = Math.min(midPoint+interval, maxDataValue);
				
				calcuateCategories(categorySlider.getValue());
				
			}
			
			// If we're selecting then make the new selection
			if (isSelecting) {
				isSelecting = false;
				if (Math.abs(selectionStart-selectionEnd) <= 3) {
					// Don't make a selection from a really small region
					return;
				}
				
				currentMinValue = getXValue(Math.min(selectionStart,selectionEnd));
				currentMaxValue = getXValue(Math.max(selectionStart,selectionEnd));

				calcuateCategories(categorySlider.getValue());

			}
			
			
		}
	
	}
	
	/**
	 * The Class HistogramCategory.
	 */
	private class HistogramCategory {
		
		/** The min value. */
		public double minValue;
		
		/** The max value. */
		public double maxValue;
		
		/** The x start. */
		public int xStart = 0;
		
		/** The x end. */
		public int xEnd = 0;
		
		/** The count. */
		public int count = 0;
		
		public float valueTotal = 0;
		
		/**
		 * Instantiates a new histogram category.
		 * 
		 * @param minValue the min value
		 * @param maxValue the max value
		 */
		public HistogramCategory (double minValue, double maxValue) {
			this.minValue = minValue;
			this.maxValue = maxValue;
		}
		
		public void addValue (float value) {
			valueTotal += value;
			count++;
		}
		
		public float value () {
			if (count > 0) {
				return valueTotal/count;
			}
			
			return 0;
		}
		
	}
	
	private class DistanceValue implements Comparable<DistanceValue> {
		
		public int distance;
		public float value;
		
		public DistanceValue (int distance, float value) {
			if (minDataValue == 0) {
				this.distance = Math.abs(distance);
			}
			else {
				this.distance = distance;
			}
			this.value = value;
		}

		public int compareTo(DistanceValue dv) {
			return distance - dv.distance;
		}
	}
	
	
	/**
	 * The Class StatusPanel.
	 */
	private class StatusPanel extends JPanel {
		
		/** The label. */
		private JLabel label;
				
		/**
		 * Instantiates a new status panel.
		 */
		public StatusPanel () {
			setBackground(Color.WHITE);
			setOpaque(true);
			label = new JLabel("No selected category",JLabel.LEFT);
			setLayout(new BorderLayout());
			add(label,BorderLayout.WEST);
			setBorder(BorderFactory.createEmptyBorder(2, 5, 2, 2));
		}
		
		/**
		 * Sets the selected category.
		 * 
		 * @param category the new selected category
		 */
		public void setSelectedCategory (HistogramCategory category) {
			if (category == null) {
				label.setText("No selected Category");
			}
			else {
				label.setText("Value Range = "+df.format(category.minValue)+" to "+df.format(category.maxValue)+" Count = "+category.value());
			}
		}
		
	}




	
}
