/**
 * Copyright 2013-15 Simon Andrews
 *
 *    This file is part of SeqMonk.
 *
 *    SeqMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SeqMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with SeqMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.babraham.SeqMonk.Displays.VennDiagram;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.image.BufferedImage;
import java.io.PrintWriter;
import java.util.Arrays;

import javax.swing.JFrame;
import javax.swing.JPanel;

import uk.ac.babraham.SeqMonk.DataTypes.Probes.Probe;
import uk.ac.babraham.SeqMonk.DataTypes.Probes.ProbeList;
import uk.ac.babraham.SeqMonk.Gradients.ColourIndexSet;
import uk.ac.babraham.SeqMonk.Utilities.ImageSaver.SVGProducer;

public class VennDiagramPanel extends JPanel implements SVGProducer {
	
	private ProbeList probeList1 = null;
	private ProbeList probeList2 = null;
	
	private int probeList1Count = 0;
	private int probeList2Count = 0;
	private int probeList12Count = 0;
	
	
	public VennDiagramPanel () {
		
	}
	
	public void setLeftProbeList (ProbeList list) {
		probeList1 = list;
		updateCounts();
	}
	
	public void setRightProbeList (ProbeList list) {
		probeList2 = list;
		updateCounts();
	}
	
	private void updateCounts () {
		
		probeList12Count = 0;
		probeList2Count = 0;
		probeList12Count = 0;
		
		if (probeList1 != null) {
			probeList1Count = probeList1.getAllProbes().length;
		}
		else {
			probeList1Count = 0;
		}

		if (probeList2 != null) {
			probeList2Count = probeList2.getAllProbes().length;
		}
		else {
			probeList2Count = 0;
		}
		
		if (probeList1 == null || probeList2 == null) {
			// Go no further
			repaint();
			return;
		}
		
		// Find the number of intersects
		
		// Mix the two lists together
		Probe [] mixedProbes = new Probe[probeList1Count+probeList2Count];
		
		Probe [] probes1 = probeList1.getAllProbes();
		Probe [] probes2 = probeList2.getAllProbes();
		
		for (int i=0;i<probes1.length;i++) {
			mixedProbes[i] = probes1[i];
		}
		
		for (int i=0;i<probes2.length;i++) {
			mixedProbes[probeList1Count+i] = probes2[i];
		}
		
		Arrays.sort(mixedProbes);
		
		for (int i=1;i<mixedProbes.length;i++) {
			if (mixedProbes[i] == mixedProbes[i-1]) {
				++probeList12Count;
				--probeList1Count;
				--probeList2Count;
			}
		}

		repaint();
	}
	
	public void paint (Graphics g) {
		
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, getWidth(),getHeight());
		
		// Create the two circle objects
		
		// Work out what diameter we can use for the circles.
		
		int heightDiameter = getHeight() - 20; // 10px border on each end
		int widthDiameter = (int)((getWidth()-20) / (3d/2));
		
		int diameter = Math.min(heightDiameter,widthDiameter);
		
		Ellipse2D.Double leftCircle = new Ellipse2D.Double(getWidth()/2 - ((diameter*3)/4),getHeight()/2 - (diameter/2),diameter,diameter);
		Ellipse2D.Double rightCircle = new Ellipse2D.Double(getWidth()/2 - (diameter/4),getHeight()/2 - (diameter/2),diameter,diameter);
		
		Graphics2D g2d = (Graphics2D)g;
		
		g2d.setColor(ColourIndexSet.getColour(0));
		g2d.fill(leftCircle);
		g.setColor(Color.BLACK);
		g2d.draw(leftCircle);
		g2d.setColor(ColourIndexSet.getColour(1));
		g2d.fill(rightCircle);
		g.setColor(Color.BLACK);
		g2d.draw(rightCircle);
		
		
		Area intesect12 = new Area(leftCircle);
		intesect12.intersect(new Area(rightCircle));
		g.setColor(ColourIndexSet.getColour(3));
		g2d.fill(intesect12);
		g.setColor(Color.BLACK);
		g2d.draw(intesect12);
		
		// No draw the numbers
		
		// The intersect number
		String intersectNumber = ""+probeList12Count;
		g.drawString(intersectNumber,getWidth()/2 - (g.getFontMetrics().stringWidth(intersectNumber))/2,getHeight()/2);

		// The left circle number
		String leftCircleNumber = ""+probeList1Count;
		g.drawString(leftCircleNumber,(getWidth()/2)-(diameter/2) - (g.getFontMetrics().stringWidth(intersectNumber))/2,getHeight()/2);

		// The right circle number
		String rightCircleNumber = ""+probeList2Count;
		g.drawString(rightCircleNumber,(getWidth()/2)+(diameter/2) - (g.getFontMetrics().stringWidth(intersectNumber))/2,getHeight()/2);

	}
	

	public static void main (String [] args) {
		JFrame frame = new JFrame();
		frame.setSize(800,600);
		frame.setContentPane(new VennDiagramPanel());
		frame.setVisible(true);
	}

	public void writeSVG(PrintWriter pr) {
		
		pr.print("<?xml version=\"1.0\" standalone=\"no\"?>\n");
		pr.print("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n");
		pr.print("<svg width=\"");
		pr.print(getWidth());
		pr.print("\" height=\"");
		pr.print(getHeight());
		pr.print("\" version=\"1.1\" xmlns=\"http://www.w3.org/2000/svg\">\n");

		BufferedImage b = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_INT_RGB);
		Graphics g = b.getGraphics();			

		// This code is adapted from the main paint method and should be updated when that is changed
		
		// Make the background white
		
		pr.print("<rect width=\"");
		pr.print(getWidth());
		pr.print("\" height=\"");
		pr.print(getHeight());
		pr.print("\"");
		pr.print(" x=\"");
		pr.print(0);
		pr.print("\" y=\"");
		pr.print(0);
		pr.print("\" style=\"fill:rgb(255,255,255");
		pr.print(");stroke:none\"/>\n");

		
		// Create the two circle objects
		
		// Work out what diameter we can use for the circles.
		
		int heightDiameter = getHeight() - 20; // 10px border on each end
		int widthDiameter = (int)((getWidth()-20) / (3d/2));
		
		int diameter = Math.min(heightDiameter,widthDiameter);

		int leftCircleCenter = getWidth()/2 - ((diameter*3)/4) + (diameter/2);
		int rightCircleCenter = leftCircleCenter+(diameter/2);

		
		// Draw the left circle		
		pr.print("<ellipse cx=\"");
		pr.print(leftCircleCenter);
		pr.print("\" cy=\"");
		pr.print(getHeight()/2);
		pr.print("\" rx=\"");
		pr.print(diameter/2);
		pr.print("\" ry=\"");
		pr.print(diameter/2);
		pr.print("\" fill-opacity=\"0.5\" style=\"fill:rgb(255,0,0);stroke:rgb(0,0,0)\"/>\n");

		// Draw the right circle		
		pr.print("<ellipse cx=\"");
		pr.print(rightCircleCenter);
		pr.print("\" cy=\"");
		pr.print(getHeight()/2);
		pr.print("\" rx=\"");
		pr.print(diameter/2);
		pr.print("\" ry=\"");
		pr.print(diameter/2);
		pr.print("\" fill-opacity=\"0.5\" style=\"fill:rgb(0,255,0);stroke:rgb(0,0,0)\"/>\n");
		
		//TODO: Work out how to draw the intersect
//		Area intesect12 = new Area(leftCircle);
//		intesect12.intersect(new Area(rightCircle));
//		g.setColor(ColourIndexSet.getColour(3));
//		g2d.fill(intesect12);
//		g.setColor(Color.BLACK);
//		g2d.draw(intesect12);
		
		// Now draw the numbers
		
		// The intersect number
		String intersectNumber = ""+probeList12Count;
		
		pr.print("<text x=\"");
		pr.print(getWidth()/2 - (g.getFontMetrics().stringWidth(intersectNumber))/2);
		pr.print("\" y=\"");
		pr.print(getHeight()/2);
		pr.print("\" fill=\"rgb(0,0,0");
		pr.print(")\" font-family=\"Arial\" font-size=\"");
		pr.print(12);
		pr.print("\">");
		pr.print(intersectNumber);
		pr.print("</text>\n");

		
		g.drawString(intersectNumber,getWidth()/2 - (g.getFontMetrics().stringWidth(intersectNumber))/2,getHeight()/2);

		// The left circle number
		String leftCircleNumber = ""+probeList1Count;

		pr.print("<text x=\"");
		pr.print((getWidth()/2)-(diameter/2) - (g.getFontMetrics().stringWidth(intersectNumber))/2);
		pr.print("\" y=\"");
		pr.print(getHeight()/2);
		pr.print("\" fill=\"rgb(0,0,0");
		pr.print(")\" font-family=\"Arial\" font-size=\"");
		pr.print(12);
		pr.print("\">");
		pr.print(leftCircleNumber);
		pr.print("</text>\n");


		// The right circle number
		String rightCircleNumber = ""+probeList2Count;

		pr.print("<text x=\"");
		pr.print((getWidth()/2)+(diameter/2) - (g.getFontMetrics().stringWidth(intersectNumber))/2);
		pr.print("\" y=\"");
		pr.print(getHeight()/2);
		pr.print("\" fill=\"rgb(0,0,0");
		pr.print(")\" font-family=\"Arial\" font-size=\"");
		pr.print(12);
		pr.print("\">");
		pr.print(rightCircleNumber);
		pr.print("</text>\n");
		
		pr.print("</svg>\n");

	}
	
}
